﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Linq;

using osm2mongo.OsmData;

using MongoDB.Driver;
using MongoDB.Driver.Builders;

namespace osm2mongo
{
	public class OsmImportManager
	{


		public string OsmFilepath { get; set; }

		public string MongoHost { get; set; }

		public string MongoDbName { get; set; }

		private MongoProvider _db;

		// Set some counters
		int _totalNodes = 0;
		int _totalWays = 0;
		int _totalRelations = 0;

		/// <summary>
		/// The list containing the processed osm objects.
		/// </summary>
		private List<IOsmEntity> _dataObjects = new List<IOsmEntity>();

		private const string NodesCollectionName = "nodes";
		private const string WaysCollectionName = "ways";
		private const string RelationsCollectionName = "relations";

		/// <summary>
		/// Create the timer to update the update the UI.
		/// </summary>
		Stopwatch _sw = new Stopwatch();

		#region Constructor

		/// <summary>
		/// Creates a new instance of the OsmMongoImporter object
		/// </summary>
		/// <param name="osmFile"></param>
		/// <param name="mongoHost"></param>
		/// <param name="mongoDbName"></param>
		public OsmImportManager()
		{

		}

		/// <summary>
		/// Creates a new instance of the OsmMongoImporter object
		/// </summary>
		/// <param name="osmFile"></param>
		/// <param name="mongoHost"></param>
		/// <param name="mongoDbName"></param>
		public OsmImportManager(string osmFile, string mongoHost, string mongoDbName)
		{

			// Set the path to the file, and the mongo database details.
			this.OsmFilepath = osmFile;

			this.MongoHost = (!String.IsNullOrEmpty(mongoHost) ? mongoHost : "localhost:27017");
			this.MongoDbName = (!String.IsNullOrEmpty(mongoDbName) ? mongoDbName : "osm");

			// Create the mongo provider
			_db = new MongoProvider(this.MongoHost, this.MongoDbName);

		}

		#endregion

		#region Import OSM data

		/// <summary>
		/// Import the OSM data into the mongo database.
		/// </summary>
		public void ImportData()
		{

			// Run the pre-import functionality.
			this.preImport();

			// Get the first 100 nodes
			IEnumerable<XElement> nodeData = this.StreamOsmDataNodes().Select(i => i).Where(i => i.Name == "node");
			IEnumerable<XElement> wayData = this.StreamOsmDataNodes().Select(i => i).Where(i => i.Name == "way");
			IEnumerable<XElement> relationData = this.StreamOsmDataNodes().Select(i => i).Where(i => i.Name == "relation");

			// Loop through the nodes
			foreach (XElement dataItem in nodeData)
			{
				// Get the attribute based on name
				processDataItem(dataItem, _totalNodes, typeof(OsmNode), NodesCollectionName);
				_totalNodes++;	
			}

			// Finish of the remain nodes in the object collection.
			batchInsertObjects(typeof(OsmNode), NodesCollectionName);

			// Loop through the ways
			foreach (XElement dataItem in wayData)
			{
				// Get the attribute based on name
				processDataItem(dataItem, _totalWays, typeof(OsmWay), WaysCollectionName);
				_totalWays++;
			}

			// Finish of the remain nodes in the object collection.
			batchInsertObjects(typeof(OsmWay), WaysCollectionName);

			// Loop through the relations
			foreach (XElement dataItem in relationData)
			{
				// Get the attribute based on name
				processDataItem(dataItem, _totalRelations, typeof(OsmRelation), RelationsCollectionName);
				_totalRelations++;
			}

			// Finish of the remain nodes in the object collection.
			batchInsertObjects(typeof(OsmRelation), RelationsCollectionName);

			// Stop the stopwatch
			_sw.Stop();

			displayOutput();

		}

		/// <summary>
		/// Processes the data item, which adds it to the collection and batch inserts after 500 milliseconds.
		/// </summary>
		/// <param name="dataItem"></param>
		/// <param name="totalCount"></param>
		/// <param name="dataType"></param>
		/// <param name="collectionName"></param>
		private void processDataItem(XElement dataItem, int totalCount, Type dataType, string collectionName)
		{
			// Reset the stop watch 
			if (totalCount == 0)
			{
				_sw.Restart();
				_dataObjects.Clear();
			}

			// If the timer has progressed 500 milliseconds, batch insert the current list of nodes.
			if (totalCount > 0 && _sw.ElapsedMilliseconds > 250)
			{
				batchInsertObjects(dataType, collectionName);

				// restart the timer
				_sw.Restart();
			}

			// Add the current node to the list for batch processing, and increment the total nodes processed.
			switch (collectionName)
			{

				case NodesCollectionName:
					_dataObjects.Add(getNodeObj(dataItem));
					break;

				case WaysCollectionName:
					_dataObjects.Add(getWayObj(dataItem));
					break;

				case RelationsCollectionName:
					_dataObjects.Add(getRelationObj(dataItem));
					break;

			}
		}

		/// <summary>
		/// Batch inserts the objects in the collection.
		/// </summary>
		/// <param name="dataType"></param>
		/// <param name="collectionName"></param>
		private void batchInsertObjects(Type dataType, string collectionName)
		{
			_db.DB.GetCollection(dataType, collectionName).InsertBatch(_dataObjects);
			_dataObjects.Clear();

			displayOutput();
		}

		#endregion

		#region Helper functions

		/// <summary>
		/// Display the progress output to the console. 
		/// </summary>
		private void displayOutput()
		{

			// Clear the current line
			Console.SetCursorPosition(0, (Console.CursorTop == 0 ? 0 : Console.CursorTop - 1));

			// Write the output to the console.
			Console.WriteLine(String.Format("Nodes processed: {0}    Ways processed: {1}    Relations processed: {2}",
				_totalNodes,
				_totalWays,
				_totalRelations));
		}

		/// <summary>
		/// This method is called prior to the data being imported, for validation and index creation purposes.
		/// </summary>
		public void preImport()
		{

			// Validate the osm file exists
			if (String.IsNullOrEmpty(this.OsmFilepath) || !File.Exists(this.OsmFilepath))
			{
				throw new IOException("The OSM data file '" + this.OsmFilepath + "' does not exist.");
			}

			// Validate the mongo server is operational. If the database does not exist, them create it. 
			prepareCollection(NodesCollectionName);
			prepareCollection(WaysCollectionName);
			prepareCollection(RelationsCollectionName);

			// Now we want to create a few indexes on the mongo database
			// Create the tag keys index
			IndexKeysBuilder keys = IndexKeys.Ascending("tagKeys");
			IndexOptionsBuilder options = IndexOptions.SetName("tagKeys_1").SetSparse(true);
			_db.DB.GetCollection<OsmNode>(NodesCollectionName).CreateIndex(keys, options);
			_db.DB.GetCollection<OsmNode>(WaysCollectionName).CreateIndex(keys, options);
			_db.DB.GetCollection<OsmNode>(RelationsCollectionName).CreateIndex(keys, options);

			// Create the location 2d geo index
			keys = IndexKeys.GeoSpatialSpherical("location");
			options = IndexOptions.SetName("location_2dgeo");
			_db.DB.GetCollection<OsmNode>(NodesCollectionName).CreateIndex(keys, options);

			Console.WriteLine("Pre-import complete. Importing data...");
			Console.WriteLine();
			Console.WriteLine();

		}

		/// <summary>
		/// Prepares the collection by dropping it and recreating it.
		/// </summary>
		/// <param name="collectionName"></param>
		private void prepareCollection(string collectionName)
		{
			// Drop and create the collection
			_db.DB.DropCollection(collectionName);
			_db.DB.CreateCollection(collectionName);
		}

		/// <summary>
		/// Gets all the node, way & relation elements from the node. 
		/// </summary>
		/// <returns></returns>
		private IEnumerable<XElement> StreamOsmDataNodes()
		{

			// Begin using the XML reader
			using (XmlReader reader = XmlReader.Create(this.OsmFilepath))
			{

				// Move the reader to the content
				reader.MoveToContent();

				// Iterate the file and get the nodes
				while (reader.Read())
				{

					switch (reader.NodeType)
					{
						// If the currently read object is an element, and it matches the node, way or relation nodes, return it. 
						case XmlNodeType.Element:
							if (reader.Name == "node" || reader.Name == "way" || reader.Name == "relation")
							{
								XElement theNode = (XElement)XElement.ReadFrom(reader);
								if (theNode != null)
								{
									yield return theNode;
								}
							}
							break;
					}
				}

			}

		}

		#endregion

		#region OSM object creators

		/// <summary>
		/// Gets a new OsmNode object from the XML object.
		/// </summary>
		/// <param name="dataItem">The XML node data.</param>
		/// <returns></returns>
		private OsmNode getNodeObj(XElement dataItem)
		{

			// Create the node
			OsmNode node = BaseOsmEntity.Create<OsmNode>(dataItem);

			// Create the float array
			float[] location = new float[2];
			location[0] = (float)dataItem.Attribute("lat");
			location[1] = (float)dataItem.Attribute("lon");
			node.Location = location;

			// return the node object
			return node;

		}

		/// <summary>
		/// Gets a new OsmWay object from the XML object.
		/// </summary>
		/// <param name="dataItem">The XML node data.</param>
		/// <returns></returns>
		private OsmWay getWayObj(XElement dataItem)
		{

			// Create the way object
			OsmWay way = BaseOsmEntity.Create<OsmWay>(dataItem);

			XName nodeName = XName.Get("nd");
			var nodes = dataItem.Descendants(nodeName);

			// Get all the tags
			foreach (XElement nodeData in nodes)
			{

				// Get the node id from the attribute
				long nodeId = (long)nodeData.Attribute("ref");

				// Add the tag to the node
				way.Nodes.Add(nodeId);

			}

			// return the way object
			return way;

		}

		/// <summary>
		/// Gets a new OsmWay object from the XML object.
		/// </summary>
		/// <param name="dataItem">The XML node data.</param>
		/// <returns></returns>
		private OsmRelation getRelationObj(XElement dataItem)
		{

			// Create the way object
			OsmRelation relation = BaseOsmEntity.Create<OsmRelation>(dataItem);

			XName memberBane = XName.Get("nd");
			var members = dataItem.Descendants(memberBane);

			// Get all the tags
			foreach (XElement memberData in members)
			{

				// Get the member data from the attributes
				string type = (string)memberData.Attribute("type");
				long refId = (long)memberData.Attribute("ref");
				string role = (memberData.Attribute("role") != null ? (string)memberData.Attribute("role") : "");

				// Add the tag to the node
				relation.Relations.Add(new OsmRelationMember()
				{
					Type = type,
					ReferenceId = refId,
					Role = role
				});

			}

			// return the way object
			return relation;

		}

		#endregion

	}
}
